/***************************
 ** Quelltext-Highlighter **
 ***************************/
// Dieses Objekt hält den Input-Quelltext (der bereits
// von < und > befreit sein muss und anwenderspezifische <>-Anmerkungen aufweisen darf)
// sowie die aktuelle Parse-Position.
// Mit zusätzlichen Rückgabewerten (etwa <hi>) ist das Funktions-Handling erleichtert.
// Die (anonymen) Funktionen bearbeiten den pre.innerHTML und ergänzen die
// Hervorhebungen für Kommentare, Schlüsselwörter, Zahlen, Strings usw.
// Deren Formatierung ist Sache von CSS.
function Parser(txt) {
 this.txt=txt;
 this.pos=0;
 this.end=function() {
  return this.pos>=this.txt.length;
 }
 this.peek=function() {		// Nächstes Zeichen lesen zum Prüfen
  return this.txt.charAt(this.pos);
 }
 this.get=function() {		// Nächstes Zeichen abholen
  return this.txt.charAt(this.pos++);
 }
// Escaptes Zeichen holen, notfalls "\r\n" oder "\n\r" (also 2 Zeichen) bei DOS-Zeilenenden
 this.getEscaped=function() {
  var ret=this.get();
  if (ret=='\r' && this.peek()=='\n') ret+=this.get();
  else if (ret=='\n' && this.peek()=='\r') ret+=this.get();
  return ret;
 }

 this.getWhite=function() {
  var ret="";
  for(;;) switch (this.peek()) {
   case ' ':
   case '\t':
   case '\r':
   case '\n':
   case '\v': ret+=this.get(); break;
   default: return ret;
  }
 }
// Präprozessor, Kommentare, Strings und reguläre Ausdrücke erkennen (C, evtl. Pascal usw.)
// Liefert null wenn <start> nicht passt
// Ist <end> gegeben wird mit <end>, dem Zeilenende oder Stringende abgebrochen
// Ist <ml> gegeben und true, wird nicht mit dem Zeilenende abgebrochen
 this.getRun=function(start,end,ml) {
  if (this.txt.substr(this.pos,start.length)!==start) return null;
  this.pos+=start.length;	// Der erste Stern muss hier verschluckt werden. (Nicht /*/)
  loop:for(;;){
   switch (this.peek()) {
    case '': break loop;	// Ende (bei Strings: Fehlerhaftes Dateiende!)
    case '\\': start+=this.get(); start+=this.getEscaped(); break;	// ggf. Zeilenfortsetzung
    case '\r':
    case '\n': if (!ml) break loop;	// Multiline-Kommentar hier nicht abbrechen!
    default: if (end && this.txt.substr(this.pos,end.length)===end) {
     start+=end;
     this.pos+=end.length;	// Ende erkannt
     break loop;
    }else start+=this.get();
   }
  }
  return start;
 }
 
 this.getToken=function() {
  this.hi="";
  var ret;
  if (ret=this.getRun("<span class=\"ln\"","</span>",true)) this.hi=null; // Zeilennummern
  else if (ret=this.getRun("<",">")) this.hi=null;	// handgemachte HTML-Zusätze
		// TODO: Zwischengeschaltete Strings können ">" enthalten!
  return ret;
 }

 this.getCToken=function(inpre) {
  var ret;
  if (ret=this.getToken()) ;
  else if (!inpre && (ret=this.getRun("#"))) this.hi="dfn";	// Präprozessor-Anweisung
  else if (inpre && (ret=this.getRun("&lt;","&gt;"))) this.hi="strong";	// #include <stdio.h> wie String
  else if (ret=this.getRun("'","'")) this.hi="strong";	// Zeichen-Konstante
  else if (ret=this.getRun('"','"')) this.hi="strong";	// String-Konstante
  else if (ret=this.getRun("/*","*/",true)) this.hi="i";// C-Kommentar
  else if (ret=this.getRun("//")) this.hi="i";		// C++-Kommentar
  else{
   var alpha=/[A-Za-z_$]/;	// Dollarzeichen für GNU-C (huch?)
   var alnum=/[0-9A-Za-z_$]/;	// für Ende des Bezeichners
   var num=/[0-9]/;
   var numal=/[0-9A-Za-z_.]/;	// Punkt (Dezimaltrennzeichen!) erlauben
   var c;
   ret=this.get();
   if (alpha.test(ret)) while (alnum.test(this.peek())) ret+=this.get();	// Bezeichner
   else if (num.test(ret)) {
    while (numal.test(this.peek())) ret+=this.get();	// Zahl
    this.hi="em";
   }
   else if (ret==='&') do ret+=c=this.get(); while(c!==';');	// &amp; &lt; &gt; als /ein/ Token
  }
  return ret;	// Ansonsten liefert diese Funktion Einzelzeichen (nicht "++", "+=" oder so)
 }

 this.getAsmToken=function() {
  var ret;
  if (ret=this.getToken()) ;	
  else if (ret=this.getRun("'","'")) this.hi="strong";	// Zeichen-Konstante
  else if (ret=this.getRun('"','"')) this.hi="strong";	// String-Konstante
  else if (ret=this.getRun("/*","*/",true)) this.hi="i";// C-Kommentar
  else if (ret=this.getRun("//")) this.hi="i";		// C++-Kommentar
  else if (ret=this.getRun(";")) this.hi="i";		// Assembler-Kommentar
  else if (ret=this.getRun("#")) this.hi="dfn";	// Präprozessor-Anweisung
  else if (ret=this.getRun("."," ")) this.hi="dfn";	// Direktive
  else{
   var alpha=/[A-Za-z_$]/;	// Dollarzeichen für GNU-C (huch?)
   var alnum=/[0-9A-Za-z_$]/;	// für Ende des Bezeichners
   var num=/[0-9]/;
   var numal=/[0-9A-Za-z_.]/;	// Punkt (Dezimaltrennzeichen!) erlauben
   var c;
   ret=this.get();
   if (alpha.test(ret)) {
    while (alnum.test(this.peek())) ret+=this.get();	// Bezeichner
    if (this.peek()==':') this.hi="label";
   }
   else if (num.test(ret)) {
    while (numal.test(this.peek())) ret+=this.get();	// Zahl
    this.hi="em";
   }
   else if (ret==='&') do ret+=c=this.get(); while(c!==';');	// &amp; &lt; &gt; als /ein/ Token
  }
  return ret;
 }
}

function hiliteC(txt,inpre) {
 var out="";
 var parser=new Parser(txt);
 var c_keywords=Array(
	"auto","break","case","char","const","continue","default","do","double","else","enum",
	"extern","float","for","goto","if","inline","int","long","register","return","short","signed","sizeof",
	"static","struct","switch","typedef","union","unsigned","void","volatile","wchar_t","while");
 var c_nonstandard=Array(
	"asm","based","_catch","cdecl","compl","_cs","_ds","_es","export","far","fastcall","_finally","_forceinline","fortran","huge","interrupt",
	"_leave","_loadds","near","_oldcall","pascal","_saveregs","_seg","_ss","stdcall","_syscall","_try","_unaligned");
 Array.prototype.contains=function(needle) {
  for (var i in this) if (this[i] == needle) return true;
  return false;
 }
 Array.prototype.contains_=function(needle) {
  for (var i in this) if (this[i] == needle || "_"+this[i] == needle) return true;
  return false;
 }
 
 while (!parser.end()) {
  out+=parser.getWhite();
  var token=parser.getCToken(inpre);
  if (parser.hi) {
   if (parser.hi=="dfn") {	// sub-tokenize!
    token=hiliteC(token,true);
   }
   token="<"+parser.hi+">"+token+"</"+parser.hi+">";	// gefundenes Highlight
  }
  else if (parser.hi==="") {	// bei <null> ist's HTML
   if (c_keywords.contains(token)) token="<b>"+token+"</b>";
   else if (c_nonstandard.contains_(token)) token="<b class=e>"+token+"</b>";
  }
  out+=token;
 }
 return out;
 // in Strings und Kommentaren könnte man noch nach URLs suchen ...
 // Sowie Funktionsreferenzen anlegen, Windows-API-Funktionen verlinken u.v.a.m.
}

function hiliteAsm(txt) {
 var out="";
 var parser=new Parser(txt);
 var avr_keywords=Array("add","adc","adiw","sub","subi","sbc","sbci","sbiw","and","andi","or","ori",
   "eor","com","neg","sbr","cbr","inc","dec","tst","clr","ser","mul","muls","mulsu","fmul","fmuls","fmulsu",
   "mov","movw","ldi","lds","ld","ldd","sts","st","std","lpm","elpm","spm","in","out","push","pop",
   "break","nop","sleep","wdr","rjmp","ijmp","eijmp","jmp","rcall","icall","eicall","call","ret","reti",
   "cpse","cp","cpc","cpi","sbrc","sbrs","sbic","sbis","brbs","brbc","breq","brne","brcs","brcc",
   "brsh","brlo","brmi","brpl","brge","brlt","brhs","brhc","brts","brtc","brvs","brvc","brie","brid",
   "sbi","cbi","lsl","lsr","rol","ror","asr","swap","bset","bclr","bst","bld",
   "sec","clc","sen","cln","sez","clz","sei","cli","ses","cls","sev","clv","set","clt","seh","clh");
 var general_keywords=Array("org","equ");
 Array.prototype.contains=function(needle) {
  for (var i in this) if (this[i]==needle) return true;
  return false;
 }

 while (!parser.end()) {
  out+=parser.getWhite();
  var token=parser.getAsmToken();
  if (parser.hi==="" && avr_keywords.contains(token.toLowerCase())) parser.hi="b";
  if (parser.hi==="" && general_keywords.contains(token.toLowerCase())) parser.hi="dfn";
  if (parser.hi) token="<"+parser.hi+">"+token+"</"+parser.hi+">";
  out+=token;
 }
 return out;
}

function hilitePre(pre) {
 var style=pre.className;
 var func=null;
 switch (style) {
  case "c": func=hiliteC; break;
  case "asm":
  case "lst": func=hiliteAsm; break;
 }
 if (func) {
  if (pre.outerHTML) pre.outerHTML=func(pre.outerHTML);	// innerHTML verliert die Newlines im IE
  else pre.innerHTML=func(pre.innerHTML);	// outerHTML geht bestimmt nicht im Linux-CHM-Viewer
 }
}

function onLoad() {
 var elems=document.body.getElementsByTagName("pre");
 for (var i=0; i<elems.length; i++) hilitePre(elems[i]);
// Als Ersatz für die fehlende Autoindizierung Bilder und Tabellen sichtbar nummerieren
 var kap=window.location.href;
 kap=kap.substring(kap.lastIndexOf("/")+1,kap.lastIndexOf("."));	// Nummer des Kapitels
 elems=document.body.getElementsByTagName("table");
 for (var i=0; i<elems.length; i++) {
  var id=elems[i].id;
  if (id.charAt(0)=="t") {
   var cap=elems[i].getElementsByTagName("caption")[0];
   if (cap) {
    cap.innerHTML="<b>Tabelle "+kap+"-"+id.substr(1)+":</b> "+cap.innerHTML;
   }
  }
  var e1=elems[i].getElementsByTagName("sup");		// Verweise zu Tabellenfußnoten automatisch generieren
  for (var j=0;j<e1.length;j++) {
   var inner=e1[j].innerHTML;
   if (inner.charAt(0)=="(" && inner.charAt(2)==")") {
    e1[j].innerHTML="<a href='#"+id+"n"+inner.charAt(1)+"'>"+e1[j].innerHTML;
   }
  }
 }
 elems=document.body.getElementsByTagName("figure");
 for (var i=0; i<elems.length; i++) {
  var id=elems[i].id;
  if (id.charAt(0)=="b") {
   var cap=elems[i].getElementsByTagName("figcaption")[0];
   if (cap) {
    cap.innerHTML="<b>Bild "+kap+"-"+id.substr(1)+":</b> "+cap.innerHTML;
   }
  }
 }
}

document.createElement("figure");
document.createElement("section");
document.createElement("figcaption");
document.createElement("aside");
if (window.addEventListener) window.addEventListener("load",onLoad);
else if (window.attachEvent) window.attachEvent("onload",onLoad);
else window.onload=onLoad;	// Handler zuweisen

/*************************
 ** Befehlsbeschreibung **
 *************************/

function flagbit(t) {
 var s="&harr;";
 if (t===null) s="-";
 else if (t===0 || t===1) s=t;
 document.write("<td align=center>"+s);
}

function flagtext(f,t,i) {
 if (t!==null) {
  switch (t) {
   case 0: t="Das "+f+"-Flag wird gelöscht."; break;
   case 1: t="Das "+f+"-Flag wird gesetzt."; break;
   case 2:
   case 3: switch (f) {
    case "H": t="Das H-Flag wird gesetzt, wenn ein Übertrag von Bit 3 auf Bit 4 erfolgte, andernfalls wird das Flag gelöscht."; break;
    case "S": t="Das S-Flag wird "+(t==3?"gleich dem N-Flag gesetzt.":"gesetzt, wenn entweder das N-Flag oder das V-Flag gesetzt ist. Wenn beide Flags gleich sind, wird das S-Flag gelöscht."); break;
    case "V": t="Das V-Flag wird gesetzt, wenn aus der Operation ein Zweierkomplement-Überlauf erfolgt, andernfalls wird das Flag gelöscht."; break;
    case "N": t="Das N-Flag wird gesetzt, wenn das Bit "+(t==3?15:7)+" des Ergebnisses gesetzt ist, andernfalls wird es gelöscht."; break;
    case "Z": t="Das Z-Flag wird gesetzt, wenn das Ergebnis "+(t==3?"0x0000":"0x00")+" ist, andernfalls wird das Flag gelöscht."; break;
    case "C": t="Das C-Flag wird gesetzt, wenn im Ergebnis ein Überlauf ins Bit "+(t==3?16:8)+" erfolgt, andernfalls wird es gelöscht."; break;
   }break;
   case 4: switch (f) {
    case "Z": t="Das Z-Flag wird gelöscht, wenn das Ergebnis ungleich 0x00 ist, sonst bleibt es unverändert."; break;
   }break;
   case 6: t="Wird mit dem Makro CL"+f+" gelöscht, ansonsten bleibt es unverändert"; break;
   case 7: t="Wird mit dem Makro SE"+f+" gesetzt, ansonsten bleibt es unverändert"; break;
   case 8: t="Wird mit dem Befehl BCLR "+i+" gelöscht, ansonsten bleibt es unverändert"; break;
   case 9: t="Wird mit dem Befehl BSET "+i+" gesetzt, ansonsten bleibt es unverändert"; break;
  }
  document.write("<tr valign=top><th>"+f+":<td>"+t);
 }
}

function BB(
 titel,		// AND — Logical AND
 syntax,	// AND Rd,Rr
 funktion,	// Führt eine logische UND-Verknüpfung aus ...
 operation,	// Rd ← Rd • Rr
 operanden,	// 0 ≤ d ≤ 31, 0 ≤ r ≤ 31
 words,		// 1
 zyklen,	// 1
 opcode,	// 0010 00rd dddd rrrr
 i,		// null
 t,		// null
 h,		// null
 s,		// "... Wenn beide Flags gleich sind, wird das S-Flag gelöscht."
 v,		// 0
 n,		// "gesetzt, wenn das MSB des Ergebnisses gesetzt ist, andernfalls wird es gelöscht."
 z,		// "..."
 c,		// "..."
 beispiel) {	// mehrzeiliger Text
 document.writeln("<h2>"+titel+"</h2>");
 document.writeln("<table class=\"leer\">");
 document.writeln("<tr valign=top><th>Syntax:<td>"+syntax);
 document.writeln("<tr valign=top><th>Funktion:<td>"+funktion);
 document.writeln("<tr valign=top><th>Operation:<td>"+operation);
 document.writeln("<tr valign=top><th>Operanden:<td>"+operanden);
 if (operation.indexOf("PC")<0) {
  document.writeln("<tr valign=top><th>Programmzähler:<td>PC ← PC + "+words);
 }
 document.writeln("<tr valign=top><th>Words:<td>"+words+" ("+2*words+" Byte)");
 document.writeln("<tr valign=top><th>Zyklen:<td>"+zyklen);
 document.writeln("<tr valign=top><th>"+16*words+"-Bit-Opcode:<td>"+opcode);
 document.writeln("</table>\n");
 document.writeln("<h3>Flags im Status-Register SREG</h3>\n");
 document.writeln("<table class=\"sreg\"><colgroup span=8 width=40>");
 document.writeln("<tr><th>I<th>T<th>H<th>S<th>V<th>N<th>Z<th>C");
 document.write("<tr>");
 flagbit(i);
 flagbit(t);
 flagbit(h);
 flagbit(s);
 flagbit(v);
 flagbit(n);
 flagbit(z);
 flagbit(c);
 document.writeln("\n</table>\n");
 if (i||t||h||s||v||n||z||c) {
  document.writeln("<table class=\"leer\" cellspacing=10 cellpadding=10>");
  flagtext("I",i,7);
  flagtext("T",t,6);
  flagtext("H",h,5);
  flagtext("S",s,4);
  flagtext("V",v,3);
  flagtext("N",n,2);
  flagtext("Z",z,1);
  flagtext("C",c,0);
  document.writeln("</table>\n");
 }
 if (beispiel) {
  document.writeln("<h3>Beispiel</h3>");
  document.writeln("<pre class=asm>");
  document.write(beispiel);
  document.writeln("</pre>");
  onLoad();	// nochmal parsen
 }
}
